#!/usr/bin/python

from distutils.core import setup

setup(
    name='catq',
    version='0.2.0',
    packages=['catq'],
    description='Simple resource query language.',
    author='Ebubekir Dirican',
    author_email='ebubekirdirican@gmail.com',
    url='https://github.com/ebbkrdrcn/catq'
)